

/*MIT License

  Copyright (c) 2020 Thomas Schucker

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/
/*original software by Thomas Schucker as noted above and detailed here:
   https://teaandtechtime.com/arduino-ble-cycling-power-service/ . Further modified by Greg Masik
   for use with Proform TDF Spin Bike Generation 2 to interact with Zwift program. (March 2021).
   Program works with Arduino Nano 33 BLE.
*/

#include "Arduino.h"
/* For the bluetooth funcionality */
#include <ArduinoBLE.h>

/* Device name which can be scene in BLE scanning software. */
#define BLE_DEVICE_NAME               "Arduino Nano 33 BLE "
/* Local name which should pop up when scanning for BLE devices. */
#define BLE_LOCAL_NAME                "Cycle Power BLE"


BLEService CyclePowerService("1818");
BLECharacteristic CyclePowerFeature("2A65", BLERead, 4);
BLECharacteristic CyclePowerMeasurement("2A63", BLERead | BLENotify, 8);
BLECharacteristic CyclePowerSensorLocation("2A5D", BLERead, 1);

unsigned char bleBuffer[8];
unsigned char slBuffer[1];
unsigned char fBuffer[4];




//Delcaring the inputs, outputs and maximum power limit
const int rpmSensIn = 3; //input from bike reed switch
const int ledPin = 13;  //will pulse once per rev of crank
const int bikeResistPin = A2; //input from bike resistance motor potentiometer
const int maxPwrCap = 500; //cap on maximum wattage to be fed to zwift

// setting of flags for service
unsigned short flags = 0x20;
byte sensorlocation = 0x0D; //from spec- rear hub

volatile bool pulseEvent = LOW;// first or second pulse per rpm-there are 2 magnets on crank
volatile bool Pulse2Status = LOW; //will be set high for second pulse, once per crank rev
unsigned long timestamp = 0;
unsigned long lastPulseTime = 0;
unsigned long newTime = 0;
unsigned long oldTime = 0;
unsigned long rotationTime = 0;
unsigned long crankEventTime = 0;
unsigned long updateInterval = 0;
unsigned long lastUpdateTime = 0;
float bikeResistVal = 0;
float resistVolts = 0;
float bikePower = 0;
float bikePowerFudge = 0.9; //Variable to adjust bike power fed to game to better match bike display. 1.0 give no adjustment
float arduinoSuppVoltage = 3.3; //Ard Uno 5.0 volts, Adrino NanoBle 3.3 volts
int debounceDelay = 10000; //microseconds
int RPM = 0;
int revCount = 0;
int CERpm = 0; // crank event based RPM
//following values to be sent to zwift
unsigned short power16bit = 0;
unsigned short revCount16bit = 0;
unsigned short crankEvent16bit = 0;

void checkRpmPulse(void); //prototype for interrupt routine on bike crank pulse


// the setup function runs once when you press reset or power the board
void setup() {
  pinMode(ledPin, OUTPUT);
  pinMode(rpmSensIn, INPUT);
  // Serial.begin(9600);  //for testing

  attachInterrupt(digitalPinToInterrupt(rpmSensIn), checkRpmPulse, RISING);//monitoring collector of transistor to buffer i/p

  // bluetooth stuff following:
  if (!BLE.begin())
  {
    while (1);
  }
  else
  {
    BLE.setDeviceName(BLE_DEVICE_NAME);
    BLE.setLocalName(BLE_LOCAL_NAME);
    BLE.setAdvertisedService(CyclePowerService);
    CyclePowerService.addCharacteristic(CyclePowerFeature);
    CyclePowerService.addCharacteristic(CyclePowerMeasurement);
    CyclePowerService.addCharacteristic(CyclePowerSensorLocation);

    BLE.addService(CyclePowerService);
    BLE.advertise();


  }
}


//interrupt routine follows
void checkRpmPulse() {
  int reading = digitalRead(rpmSensIn);
  if (reading == HIGH)
  {
    delayMicroseconds(debounceDelay);
    reading = digitalRead(rpmSensIn);
    if (reading == HIGH)
    {
      pulseEvent = HIGH;
      Pulse2Status = !Pulse2Status;
    }
  }


}



// the loop function runs over and over again forever
void loop()
{
  // put your main code here, to run repeatedly:
  BLEDevice central = BLE.central();
  if (central)
  {
    /*
      If a BLE device is connected, power data will
      calculated from bike magnet position voltage and crank rev pulses.
    */
    while (central.connected())
    {
      if (pulseEvent == HIGH && Pulse2Status == HIGH) {

        timestamp = millis();
        revCount = revCount + 1;
        newTime = millis();
        rotationTime = newTime - oldTime;
        oldTime = newTime;
        //crankEventTime = (rotationTime * 0.976);
        RPM = 60000 / rotationTime;
        //CERpm = 60000/crankEventTime;
        //crankEventTime = millis() * 0.9765625;//reduction due to reed closure not required
        crankEventTime = millis();
        pulseEvent = LOW;
      }

      digitalWrite(ledPin, pulseEvent);
      bikeResistVal = analogRead(bikeResistPin);
      resistVolts = ((bikeResistVal * arduinoSuppVoltage) / 1024);
      //Serial output for testing
      //Serial.print("Resist Value =");
      //Serial.println(bikeResistVal);
      //Serial.print("resist volts = ");
      //Serial.println(resistVolts);

      bikePower = (((9.66 * RPM) - (264)) * (resistVolts * resistVolts) + ((((-14.16) * RPM) + 417) * resistVolts) + ((6.27 * RPM) - 196));

      bikePower = bikePower * bikePowerFudge; //added Dec 18, 2021 to allow tuning of bike power fed to game
      if (bikePower < 15 || RPM < 10 || ((millis() - timestamp) > 6000))
        // clamps lower limit of power/rpm
        //ensures that if no pulse in 6 seconds zero power and rpm reported
      {
        bikePower = 0.0;
        RPM = 0;
      }

      if (bikePower > maxPwrCap)
        //upper power limit to be reported
      {
        bikePower = maxPwrCap;
      }

      //following formats data for transmission buffers

      crankEvent16bit = (short int)crankEventTime;
      revCount16bit = (short int)revCount;
      power16bit = (short int)bikePower;

      /* testing stuff follows
          Serial.print("Power short =");
        Serial.println(power16bit);
        Serial.print("RPM= ");
        Serial.println(RPM);
        Serial.print("short revCount ");
        Serial.println(revCount16bit);
        Serial.print("short crankeventtime= ");
        Serial.println(crankEvent16bit);*/

      updateInterval = millis() - lastUpdateTime; //limit service updates to every 500ms
      if (updateInterval > 500)
      {
        bleBuffer[0] = flags & 0xff;
        bleBuffer[1] = (flags >> 8) & 0xff;
        bleBuffer[2] = power16bit & 0xff;
        bleBuffer[3] = (power16bit >> 8) & 0xff;
        bleBuffer[4] = revCount16bit & 0xff;
        bleBuffer[5] = (revCount16bit >> 8) & 0xff;
        bleBuffer[6] = crankEvent16bit & 0xff;
        bleBuffer[7] = (crankEvent16bit >> 8) & 0xff;

        slBuffer[0] = sensorlocation & 0xff;

        fBuffer[0] = 0x00;
        fBuffer[1] = 0x00;
        fBuffer[2] = 0x00;
        fBuffer[3] = 0x08;

        CyclePowerFeature.writeValue(fBuffer, 4);
        CyclePowerMeasurement.writeValue(bleBuffer, 8);
        CyclePowerSensorLocation.writeValue(slBuffer, 1);
        lastUpdateTime = millis();
      }
    }

  }
}
